#ifndef __AVRSPX_H
#define __AVRSPX_H

#ifdef _WIN32
#include <windows.h>
#endif
//----------------------------------------------------------------------------
#ifdef __BORLANDC__
# pragma warn -8012
# pragma warn -8060
# pragma warn -8065
#endif

#ifdef _MSC_VER
# pragma warning(disable : 4244)
# pragma warning(disable : 4267)
# pragma warning(disable : 4996)
#endif

#ifndef bool
# define bool	char
# define true	1
# define false	0
#endif

#ifndef _WINDEF_
typedef unsigned int    BOOL;
typedef unsigned char   BYTE;
typedef unsigned short	WORD;
typedef unsigned long	DWORD;
#define TRUE  1
#define FALSE 0
#endif /* _WINDEF_ */

#ifdef F_LOCK // !!! for GCC
#undef F_LOCK
#endif
 
#define MESS(str)	fputs(str, stderr)
//#define MESS(...)	fprintf(stderr, __VA_ARGS__)
#define FUSEFILE "fuse.txt"
//#define INIFILE "avrsp.ini"


/* Program error codes */

#define	RC_FAIL		1
#define	RC_FILE		2
#define	RC_INIT		3
#define RC_DEV		4
#define	RC_SYNTAX	5


/* Buffer size for flash/eeprom */

#define	MAX_FLASH	(256*1024)	/* Max Flash size */
#define	MAX_EEPROM	(  4*1024)	/* Max EEPROM size */
#define	PIPE_WINDOW	32			/* Pipe window for SPI bridge */


/* Device property structure */

#define N0000       0x000000
#define L0000       0x000102
#define AT90S1200   0x1E9001
#define TINY12      0x1E9005

typedef struct _DEVPROP {
	char	*Name;			/* Device name */
	DWORD	Signature;		/* Device signature bytes */
	DWORD	FlashSize;		/* Flash memory size in unit of byte */
	WORD	FlashPage;		/* Flash page size (0 is byte-by-byte) */
	DWORD	EepromSize;		/* EEPROM size in unit of byte */
	WORD	EepromPage;		/* EEPROM page size (0 is byte-by-byte) */
	WORD	FlashWait;		/* Wait time for flash write */
	WORD	EepromWait;		/* Wait time for EEPROM write */
	BYTE	PollData;		/* Polling data value */
	BYTE	LockData;		/* Default lock byte (program LB1 and LB2) */
	char	FuseType;		/* Device specific fuse type */
	char	Cals;			/* Number of calibration bytes */
	BYTE	FuseMask[3];	/* Valid fuse bit mask [low, high, ext] */
} DEVPROP;

extern const DEVPROP DevLst[];	/* Device property list */



/* Device programming commands */

#define C_EN_PRG1	0xAC
#define C_EN_PRG2	0x53
#define C_ERASE1	0xAC
#define C_ERASE2	0x80
#define C_LD_ADRX	0x4D
#define C_WR_PRGL	0x40
#define C_WR_PRGH	0x48
#define C_WR_PAGE	0x4C
#define C_RD_PRGL	0x20
#define C_RD_PRGH	0x28
#define C_WR_EEP	0xC0
#define C_RD_EEP	0xA0
#define C_WR_FLB	0xAC
#define C_WR_FLBL	0xA0
#define C_WR_FLBH	0xA8
#define C_WR_FLBX	0xA4
#define C_WR_FLBK	0xE0
#define C_RD_FLB1	0x58
#define C_RD_FB1	0x50
#define C_RD_SIG	0x30
#define C_RD_CAL	0x38


/* Byte read/write ID */

#define FLASH		0
#define FLASH_NS	1
#define EEPROM		2
#define SIGNATURE	3


/* Fuse write ID */

#define F_LOW		0
#define	F_HIGH		1
#define	F_EXTEND	2
#define	F_LOCK		3


/* spi_rcvr() argument */

#define	RM_SYNC		0
#define	RM_ASYNC	1


/* Physical port properties */

typedef struct {
	WORD	PortClass;		/* Port class */
	WORD	PortNum;		/* Port number (1..)  */
//	DWORD	Baud;			/* Baud rate (for SPI bridge) */
	int		Baud;			/* Baud rate (for SPI bridge) */
//	WORD	Delay;			/* I/O delay */
	int		Delay;			/*@@@ I/O delay by t.k */
	char    *SerialNumber;	/*@@@ usbasp serial no. by t.k */
	char	*Info1, *Info2;	/* Information strings, returned by open_ifport() */
} PORTPROP;

#define DEFAULT_DELAY_VALUE	3		//@@@ by t.k
#define DEFAULT_BAUDRATE	115200	//@@@ by t.k

enum _portclass {	/* Port class */
	TY_LPT, TY_COMM, TY_VCOM, TY_BRIDGE, TY_AVRSP,
    TY_STK200, TY_XILINX, TY_LATTICE, TY_ALTERA,
    TY_USBASP, TY_RSCR
};




/* Prototypes for hardware control functions */

int open_ifport (PORTPROP *);
void close_ifport ();
void spi_reset ();
void spi_clk ();
void spi_xmit (BYTE);
BYTE spi_rcvr (BYTE);
void spi_delayedget (BYTE *, DWORD);
void delay_ms (WORD);
int spi_flush ();
FILE *open_cfgfile(char *);

//----------------------------------------------------------------------------
// by t.k
// avrspx.c

// hwctrl.c
void spi_transmit(BYTE cmd1, BYTE cmd2,BYTE cmd3, BYTE cmd4);
int spi_transmit_R(BYTE cmd1, BYTE cmd2,BYTE cmd3, BYTE mode);

// utils.c
extern char progname[];
extern char inifilename[];
#undef INIFILE
#define INIFILE inifilename

unsigned long strtoul_ex(const char *s, char **endptr, int base);
int setup_commands_ex(int argc, char **argv);
DEVPROP *search_device(char *Name);
DEVPROP *get_device(DWORD signature);
#ifndef USB_ONLY
int get_outque_count(HANDLE hComm);
#endif
void dump_port_list(void);

//----------------------------------------------------------------------------
// by t.k
// wronly.c
extern bool  f_write_only_programmer;
extern int   cmdbuf_count;
extern BYTE *cmdbuf;
int wronly_universal_command(BYTE cmd1, BYTE cmd2,BYTE cmd3, BYTE cmd4);
void wronly_programmer_init(void);
void wronly_programmer_close(void);

#ifndef USB_ONLY
// rscr.c
int rscr_setup(HANDLE h, int baud);
void rscr_close(void);
void rscr_delay_ms(WORD dly);
#endif

//----------------------------------------------------------------------------
#endif	// end of __AVRSPX_H
