/*
  isp.h - part of USBasp

  Autor..........: Thomas Fischl <tfischl@gmx.de> 
  Description....: Provides functions for communication/programming
                   over ISP interface
  Licence........: Free under certain conditions. See Documentation.
  Creation Date..: 2005-02-23
  Last change....: 2005-04-20
*/

#ifndef __isp_h_included__
#define	__isp_h_included__

#ifndef uchar
#define	uchar	unsigned char
#endif

#define	ISP_OUT   PORTB
#define ISP_IN    PINB
#define ISP_DDR   DDRB
#define ISP_RST   PB2
#define ISP_MOSI  PB3
#define ISP_MISO  PB4
#define ISP_SCK   PB5

#define ISP_DELAY		1
#define ISP_SCK_SLOW	(6+5)	//@@x 60us <- 0
#define ISP_SCK_FAST	3		//@@x      <- 1

/* Prepare connection to target device */
void ispConnect(void);

/* Close connection to target device */
void ispDisconnect(void);

/* read an write a byte from isp using software (slow) */
uchar ispTransmit_sw(uchar send_byte);

/* read an write a byte from isp using hardware (fast) */
uchar ispTransmit_hw(uchar send_byte);

/* enter programming mode */
uchar ispEnterProgrammingMode(void);

/* read byte from eeprom at given address */
uchar ispReadEEPROM(unsigned int address);

/* write byte to flash at given address */
uchar ispWriteFlash(unsigned int address, uchar data, uchar pollmode);

uchar ispFlushPage(unsigned int address, uchar pollvalue);

/* read byte from flash at given address */
uchar ispReadFlash(unsigned int address);

/* write byte to eeprom at given address */
uchar ispWriteEEPROM(unsigned int address, uchar data);

/* pointer to sw or hw transmit function */
//@@- uchar (*ispTransmit)(uchar);
uchar ispTransmit(uchar send_byte);				//@@+ by t.k

/* set SCK speed. call before ispConnect! */
void ispSetSCKOption(uchar sckoption);			//@@x by t.k

//----------------------------------------------------------------------------
//@@+ by t.k

void ispLoadExtendedAddress(uchar eaddr);	// for mage256x

#if defined(__AVR_ATmega8__)
  #define ISP_FLAGS		TWBR
#else
  #define ISP_FLAGS		GPIOR0
#endif

#define FLAG_ADDR_16	0
#define FLAG_IS_HW_SPI	7

#define is_hw_spi()		(ISP_FLAGS & _BV(FLAG_IS_HW_SPI))
#define set_hw_spi()	(ISP_FLAGS |= _BV(FLAG_IS_HW_SPI))
#define clr_hw_spi()	(ISP_FLAGS &= ~_BV(FLAG_IS_HW_SPI))
#define set_sw_spi()	clr_hw_spi()

//@@+ for mega128x/mega256x
#define set_addr16()	(ISP_FLAGS |= _BV(FLAG_ADDR_16))
#define clr_addr16()	(ISP_FLAGS &= ~_BV(FLAG_ADDR_16))
#define tst_addr16()	(ISP_FLAGS & _BV(FLAG_ADDR_16))
//----------------------------------------------------------------------------

#endif /* __isp_h_included__ */
